# Generated by Django 2.2.17 on 2020-12-10 15:04
import django.contrib.postgres.fields.jsonb
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('ansible', '0028_collectionversion_namespace_length'),
    ]

    operations = [
        migrations.AddField(
            model_name='collectionversion',
            name='files',
            field=django.contrib.postgres.fields.jsonb.JSONField(default=dict, editable=False),
        ),
        migrations.AddField(
            model_name='collectionversion',
            name='manifest',
            field=django.contrib.postgres.fields.jsonb.JSONField(default=dict, editable=False),
        ),
    ]
