# Generated by Django 2.2.16 on 2020-10-26 12:57

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('ansible', '0024_remove_collectionversion_certification'),
    ]

    operations = [
        migrations.AlterField(
            model_name='collectionversion',
            name='version',
            field=models.CharField(editable=False, max_length=128),
        ),
    ]
