# Generated by Django 2.2.4 on 2019-08-14 21:08

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('ansible', '0007_collectionversion_is_certified'),
    ]

    operations = [
        migrations.AddField(
            model_name='collectionremote',
            name='requirements_file',
            field=models.TextField(max_length=255, null=True),
        ),
    ]
